/* This file contains all the routines to go from
   different forms of house codes and units
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "cm11.h"

static char *gbl_alias[100];
static char *gbl_target[100];
static int gbl_aliasMax=0;


char gHouseChr[] = "MECKOGAINFDLPHBJ";

static X10_UNIT deviceCode[]={0x6,0xe,0x2,0xa,0x1,0x9,0x5,0xd,
                              0x7,0xf,0x3,0xb,0x0,0x8,0x4,0xc};

int deviceInt[] = {13, 5, 3, 11, 15, 7, 1, 9,
                   14, 6, 4, 12, 16, 8, 2, 10};


// Forward declarations
static int Saila(char *str);
static int Alias(char *str,int len);

static int HexChar2Nybble(char c)
{
    if (c >= 'a' && c <= 'f') return (c - 'a' + 10);
    if (c >= 'A' && c <= 'F') return (c - 'A' + 10);
    if (c >= '0' && c <= '9') return (c - '0');
    return -1;
}

int Hex2Bytes(char * inStr, BYTE * outBuf, int outSize)
{
    int b;
    int outCount = 0;
    
    while (outSize && (b=HexChar2Nybble(*inStr++)) > -1) {
         *outBuf++ = (b << 4) | HexChar2Nybble(*inStr++);
         ++outCount;
    }
    return outCount;
}


char * SpaceDrop(char * str)
/* Returns pointer to first non-whitespace character of str */
{
  char *ptr=str;
  while (*ptr<=' ' && *ptr!='\0') ++ptr;
  return ptr;
}

char *NotSpaceDrop(char *str)
/* Returns pointer to first whitespace character of str */
{
  char *ptr=str;
  while (*ptr>' ') ++ptr;
  return ptr;
}

X10_UNIT ChrToHouseCode(char chr)
/* Single character -> housecode */
{
    if (chr>='a' && chr <='p') chr += 'A' - 'a';
    if (chr>='A' && chr <='P') return deviceCode[chr-'A'] << 4;
    return -1;
}

X10_UNIT IntToDeviceCode(int device)
{
    if (device < 1 || device > 16) {
        return -1;
    }
    else return deviceCode[device-1];
}

char *UnitToStr(X10_UNIT unit, char *str, int len) 
// Converts single unit into ASCII string
{
    char s[5];
    int a;
    
    sprintf(s, "%c%d",gHouseChr[unit >> 4], deviceInt[unit & 0x0f]);
    a = Saila(s);
    strncpy(str, a > -1 ? gbl_alias[a] : s, len);
    return str;
}

    
char *UnitsToStr(X10_UNIT house, X10_UNIT_BITMAP codeMap, char *str, int len)
/* expands unit bits into ascii string of numbers delimited by commas */
{
    char s[5];
    int i, a;
    int first=1;
    char houseChr=gHouseChr[house >> 4];

    str[0]='\0';
    for (i=0; i<16; ++i) {
        X10_UNIT code = deviceCode[i];
        if (codeMap & (1<<code)) {
            sprintf(s,"%c%d",houseChr,i+1);
            a = Saila(s);
            if (!first) strncat(str,",", len-strlen(str));
            strncat(str, a>-1 ? gbl_alias[a]:s, len - strlen(str));
            first=0;
        }
    }
    return str;
}

struct AliasList {
    struct AliasList *mama;
    int index;
};

static int StrToUnitsRecurse(char *token,
                             X10_UNIT *house,
                             X10_UNIT_BITMAP *devices,
                             struct AliasList *al)
{
    char *comma;
    long tempRetv;
    int t;
    X10_UNIT tempHouse,tempUnit;
    struct AliasList newAl;

    if (DEBUG_ALIAS)
        fprintf(stderr," Recursing StrToUnits %s/%x/%x\n",
                token, *house, *devices);

    while (*token) {
        comma=strchr(token,',');
        if ((t=Alias(token,
                     comma==NULL  ? strlen(token): (comma-token))) > -1) {
            if (DEBUG_ALIAS)
                fprintf(stderr, "Found Alias %d\n", t);

            // Alias found
            for (newAl.mama=al; newAl.mama!=NULL; newAl.mama=newAl.mama->mama) {
                
                if (newAl.mama->index==t) {
                    fprintf(stderr,"Loopy alias %s\n",token);
                    return -4; /* Loopy recursion? */
                }
            }
            newAl.mama=al; newAl.index=t; /* store current alias index */
            // Now decode the alias (recursive)
            if ((tempRetv=StrToUnitsRecurse(gbl_target[t],
                                            house,
                                            devices,
                                            &newAl))<0) return tempRetv;
        } else {
            // Not an alias - try direct "Cnn" type decode 
            // (also allow form of C1,2,3...)
            if (DEBUG_ALIAS)
                fprintf(stderr," Not the alias\n");

            if ((tempHouse=ChrToHouseCode(*token)) <0) tempHouse=*house;
            else ++token;

            t = atoi(token);
            if (t<1 || t>16) {
                fprintf(stderr,"Unit %s out of range\n",token);
                return -1;
            }
            tempUnit = deviceCode[t-1];

            if (*house==-1) {
                // Store first unit
                *house=tempHouse | tempUnit;
            } else {
                if ((tempHouse ^ *house) & 0xf0) {
                    fprintf(stderr,"Housecode mismatch%02x!=%02x",
                            *house,tempHouse);
                    return -2;
                }
            }
            *devices |= (1 << tempUnit);
        }
        if (comma != NULL) {
            token = comma+1;
        }
        else token += strlen(token);
    }
    if (house<0) return -3L;
    return 0;
}

X10_UNIT StrToUnit(char *str)
{
    X10_UNIT unit = -1;
    X10_UNIT_BITMAP bitmap = 0;

    StrToUnitsRecurse(str, &unit, &bitmap, NULL);
    return unit;
}

int StrToUnits(char *str, X10_UNIT *house, X10_UNIT_BITMAP *deviceBitmap)
/* Make str into a bit mask of active units, return house code in house.
   Errors -1 = invalid number  -2=housecodes don't match
	  -3 = no housecode specified
	  -4 = Loopy recursion
   Recursively de-aliases.
*/
{
    *house = -1;
    *deviceBitmap = 0;
    if (DEBUG_ALIAS)
        fprintf(stderr,"StrToUnits %s\n", str);

    return StrToUnitsRecurse(str, house, deviceBitmap, NULL);
}

void CleanupAliasStuff()
{
    while (gbl_aliasMax) {
        free(gbl_alias[--gbl_aliasMax]);
    }
}

int AddAlias(char *abuf, char *tbuf)
{
    char *t;
    
    if ((t=malloc(strlen(abuf)+strlen(tbuf)+2))==NULL) {
	fprintf(stderr,"Error allocating alias\n");
        return -1;
    }
    strcpy(t,abuf);
    gbl_alias[gbl_aliasMax]=t;
    t += strlen(abuf) + 1;
    strcpy(t,tbuf);
    gbl_target[gbl_aliasMax++]=t;
    return gbl_aliasMax;
}

int DumpAliases(FILE *outF)
// Stuff the aliases in a file
{
    int i;
    for (i= 0; i < gbl_aliasMax; i++) {
        fprintf(outF, "ALIAS \"%s\" \"%s\"\n",
                gbl_alias[i], gbl_target[i]);
    }
    return gbl_aliasMax;
}

#ifdef NOTDEF
int ReadAliasFile(char *file)
/* Add file of aliases to the global list */
{
    FILE *inF;
    BYTE buf[132];
    char abuf[32],tbuf[132];

    if (file[0]=='-') return 0;
    
    if ((inF=fopen(file,"r"))==NULL) {
        fprintf(stderr,"Error opening alias file %s\n",file);
        return -1;
    }
    
    while (fgets(buf,132,inF)!=NULL) {
        if (buf[0]!='#' && buf[0] !='\n') {
            if (sscanf(buf,"%s %s",abuf,tbuf) == 2) {
                if (AddAlias(abuf, tbuf) < 0) {
                    fclose(inF);
                    return -1;
                }
            }
        }
    }
    fclose(inF);
    return gbl_aliasMax;
}
#endif

void FreeAlias(void)
{
    while (gbl_aliasMax > 0) {
        free(gbl_alias[gbl_aliasMax--]);
    }
}

static int Alias(char *str,int len)
/* Look for match to string, first <len> characters.  Returns index
   into Alias array, or -1 if no match. */
{
  int i;
  for (i=0; i<gbl_aliasMax; ++i) if (!strncasecmp(gbl_alias[i],str,len))
    return i;
  return -1;
}

/* Alias translation */

static int Saila(char *str)
/* Look for match to target unit (eg B3).  For disassembling. */
{
  int i;
  for (i=0; i< gbl_aliasMax; ++i) {
    if (!strcasecmp(gbl_target[i],str)) {
        return i;
    }
    
  }
  return -1;
}



