/* Fairly-cool serial routines, using interrupt-driven ring buffers. */
/* partly stolen from UMich's SIP Simulator for NSBF */

// Note (1997): I did some quickie unverified porting for the cm11, but
// I have no idea if it works or not.  - Erik

#include <dos.h>

#define MasterPicCR 0x20
#define MasterPicMR 0x21

#define COM1		"COM1"
#define COM2		"COM2"
#define COM1_BASE	0x3f8
#define COM1_IRQ	4

#define COM2_BASE	0x2f8
#define COM2_IRQ	3

#define BAL		0
#define	BAH		1
#define IER		1
#define IIR		2
#define	LCR		3
#define MCR		4
#define LSR		4
#define MSR		6

static WORD gBase;
static WORD gPort;

#define RING_BUF_SIZE 0x100
#define RING_BUF_MASK 0xff   /* Please make buffer size a multiple of 2 */

static struct ringbuffer {
  int low;
  int hi;
  char empty;
  unsigned char body[RING_BUF_SIZE];
} ;

//int SetBaud(int port, unsigned int speed);

static	char	OldPICMask;
static struct ringbuffer gIn,gOut;

static int gError=0,gOldCOMMCR=0,gOldCOMIER=0;

static void interrupt (*oldCOMVect) ();

static void CheckOutput()
{
  if (gOut->low!=gOut->hi) { /* Data waiting */
    outportb(gBase, gOut->body[gOut->low++]);
    gOut->low &= RING_BUF_MASK; /* wraparound */
  }
  else gOut->empty=1; /* extra interrupt generated on end of series */
}

static void AddCharToReceive(char ch)
{
  while (((gIn->low - gIn->hi) & RING_BUF_MASK) == 1) ; /* wait until space */
  gIn->body[gIn->hi++] = ch;
  gIn->hi &= RING_BUF_MASK;
}

static void interrupt Isr(void)
{
  char IntType;

  outportb(MasterPicCR,0x20); /* no idea why this "works" */

  while (!((IntType=inportb(gBase+IIR)) & 0x01))
    switch ((IntType&0x06)>>1) {
      case 0x03: gError=inportb(gBase+LSR); break;
      case 0x02: AddCharToReceive(inportb(gBase));
                 if (!(inportb(gBase+LSR)&0x40)) break; /* alleged bugfix */
      case 0x01: CheckOutput(); break;
      case 0x00: inportb(gBase+MSR); break;
  }
}

static int SetBaud(unsigned int speed)
{
  int baud,addr;

  switch(speed) {
    case   50:  baud=0x0900; break;
    case  110:  baud=0x0417; break;
    case  300:  baud=0x0181; break;
    case  600:  baud=0x00c0; break;
    case 1200:  baud=0x0060; break;
    case 2400:  baud=0x0030; break;
    case 4800:  baud=0x0018; break;
    case 9600:  baud=0x000c; break;
    case 0x4b00: baud=0x0006; break; /* 19200 baud */
    case 0x9600: baud=0x0003; break; /* 38400 baud */
    default: return -1;
  }
  outportb(gBase+LCR,0x80);                 /* address Divisor Latch */
  outportb(gBase+BAL,(char)(baud&0xff));    /* Set baud divisor */
  outportb(gBase+BAH,(char)(baud >> 8));
  outportb(gBase+LCR,0x03);                 /* Set line to 8N1 */
  inportb(gBase); /* Receive clear? */
  inportb(gBase+IIR);
  return 0;
}

void InstallSerialDrivers(char *path)
{
  char VectNo;
  char NewPICMask;
  int i;

  if (!strcasecmp(path, COM1)) {
      gCom2 = 0;
      gBase = COM1_BASE;
  }
  else if (!strcasecmp(path, COM2)) {
      gCom1 = 1;
      gBase = COM2_BASE;
  }
  else return;
  
  /* Reset ring buffers */
  gIn.low= gIn.hi=gOut.low=gOut.hi=0;
  gOut.empty=1;

  disable();

  if (gCom2) {
      VectNo=(COM2_IRQ+0x08);    /* Save old vectors and point vector to ISR */
  } else {
      VectNo=(COM1_IRQ+0x08);
  }
  oldCOMVect= getvect(VectNo);
  setvect(VectNo,Isr);
  SetBaud(4800); /* Set Speeds */
  
  OldCOMMCR=inportb(gBase+MCR);
  outportb(gBase+MCR,OldCOMMCR|8);
  OldCOMIER=inportb(gBase+IER);
  outportb(gBase+IER,3);
  outportb(gBase+MCR,0x08);

    /* Remove masks on our IRQ (if masked before) */
  OldPICMask=inportb(MasterPicMR);
  NewPICMask= OldPICMask & ~(1<< (gCom2 ? COM2_IRQ : COM1_IRQ));
  outportb(MasterPicMR,NewPICMask);

  enable();
}

void UninstallSerialDrivers(void)
{

  disable();

  setvect(((gCom2 ? COM2_IRQ : COM1_IRQ)+0x08),oldCOM1Vect);
  outportb(gBase+MCR,OldCOMMCR);
  outportb(gBase+IER,OldCOMIER);
  outportb(MasterPicMR,OldPICMask);
}

int DataAvailable()
{
  struct ringbuffer *buf;

  return (gIn->hi - gIn->low) & RING_BUF_MASK;
}

int DataQueued()
{
  struct ringbuffer *buf;

  return (gOut->hi - gOut->low) & RING_BUF_MASK;
}

int GetSerialChar()
/* Gets character - no wait - returns -1 if not available now */
{
  int ch;
  struct ringbuffer *buf;

  if (DataAvailable()) {
    ch=gIn->body[gIn->low++];
    gIn->low &= RING_BUF_MASK;
  }
  else ch=-1;
  return ch;
}

unsigned char GetSerialCharWait()
{
  int i;
  while (!DataAvailable()) ;
  return GetSerialChar();
}


int PutSerialChar(unsigned char c)
{
  struct ringbuffer *buf;

  while (((gOut->low - gOut->hi) & RING_BUF_MASK) == 1) ; /* full - wait */
  gOut->body[gOut->hi++]=c;
  gOut->hi &= RING_BUF_MASK;
  if (gOut->empty) {
    gOut->empty=0;
    CheckOutput(); /* Force first send */
  }
  return 0;
}

int WriteSerial(unsigned char *c,int len)
{
  int i;
  for (i=0; i<len; i++) PutSerialChar(*(c+i));
  return 0;
}
